/* Copyright (c) 2018 -- 2022
   Charles Jordan (Hokkaido University, Sapporo, Japan),
   Michael Joswig, Lars Kastner (Technische Universitaet Berlin, Germany)

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
-----------------------------------------------------------------------------
*/

///////////////////////////////////////////////////////////////////////////////
// RegularityCheckCache.hh
// 
// This file contains a cache for the regularity checks. When we walk down a
// flip and then check all the flips of the target triangulations for
// regularity, many of them will already have been previously checked, so a
// cache should be useful.
///////////////////////////////////////////////////////////////////////////////
#pragma once

#include "mptopcom_typedefs.hh"
#include "LRUCache.hh"
#include "GKZ.hh"
// #include "TriangNode.hh"
#include "TriangFlips.hh"
#include <random>


#include <unordered_set>
#include <polymake/Main.h>
#include <polymake/Matrix.h>
#include <polymake/linalg.h>
#include <polymake/Set.h>
#include <polymake/SparseVector.h>
#include <polymake/polytope/is_regular.h>
#include <polymake/polytope/convex_hull.h>
#include <polymake/RandomGenerators.h>


#include <polymake/polytope/solve_LP.h>


namespace pm {

   template <>
      struct iterator_cross_const_helper<__sc_const_iterator<IntegerSet>, true> {
         typedef __sc_const_iterator<IntegerSet> iterator;
         typedef __sc_const_iterator<IntegerSet> const_iterator;
      };

}

namespace std {
   template <>
      struct hash<std::pair<TriangNode, Flip>> {
         std::size_t operator()(const std::pair<TriangNode, Flip>& key) const {
            return std::hash<TriangNode>()(key.first) + std::hash<Flip>()(key.second);
         }
      };
};

struct Subsystem {
   pm::Set<Int> active_rows, rays, non_rays, undecided;
   Subsystem() {}
   Subsystem(const Subsystem& sin){
      active_rows = sin.active_rows;
      rays = sin.rays;
      non_rays = sin.non_rays;
      undecided = sin.undecided;
   }
   void print() const{
      pm::cerr << "AR: " << active_rows;
      pm::cerr << " R: " << rays;
      pm::cerr << " NR: " << non_rays;
      pm::cerr << " U: " << undecided;
      pm::cerr << std::endl;
   }
};


template<typename FlipInfoType>
class HeuristicSolver {
   private:
      const TriangNode& source;
      const gkz_type& gkz;
      std::unordered_map<Flip, FlipInfoType>& fimap;
      pm::SparseMatrix<pm::Rational> flip_gkz_vectors, augmented_flip_gkz_vectors;
      pm::Set<Int> regular, irrelevant, non_regular, no_lp_necessary;
      pm::Set<pm::Set<Int>> groups;
      Int lp_necessary;
      
      const gkz_vector_type& get_gkz_inner(const Flip& flip, FlipInfoType& fi) const {
         if(!fi.gkz_computed){
            fi.flip_gkz = gkz.evaluate(flip);
            fi.gkz_computed = true;
         }
         return fi.flip_gkz;
      }
      

      void update_config(Subsystem& config, Subsystem& subconfig){
         config.rays += subconfig.rays;
         config.non_rays += subconfig.non_rays;
         config.undecided -= subconfig.rays;
         config.undecided -= subconfig.non_rays;
         config.active_rows -= subconfig.non_rays;
      }

      void recursive_process(Subsystem& config) {
         // config.print();
         if(config.undecided.size() == 0) return;
         if(config.undecided.size() == 1 && config.active_rows.size() == 2) {
            // Oddly specific case.
            bool check = pm::rank(flip_gkz_vectors.minor(config.active_rows, pm::All)) == 2;
            if(check) config.rays += config.undecided;
            else {
               config.non_rays += config.undecided;
               config.active_rows -= config.undecided;
               return;
            }
         }
         Int i = 0;
         for(const auto& c:cols(flip_gkz_vectors)){
            pm::Set<Int> check(support(c) * config.active_rows);
            if(check.size() == 0) continue;
            else if(check.size() == 1){
               config.rays += check;
               config.undecided -= check;
               config.active_rows -= check;
               if(config.undecided.size() > 0){
                  Subsystem subconfig(config);
                  recursive_process(subconfig);
                  update_config(config, subconfig);
               }
            } else {
               pm::Set<Int> positives, negatives;
               for(const auto& e: check){
                  if(c[e] > 0) positives += e;
                  if(c[e] < 0) negatives += e;
               }
               if(positives.size() == 1 || negatives.size() == 1){
                  pm::Set<Int> eliminator = positives.size() == 1 ? positives : negatives;
                  Int elim = eliminator.front();
                  pm::Set<Int> eliminatee = positives.size() == 1 ? negatives : positives;
                  config.rays += elim;
                  config.undecided -= elim;
                  if(eliminatee.size() == 1){ 
                     config.rays += eliminatee;
                     config.undecided -= eliminatee;
                  }
                  if(config.undecided.size() > 0 && ((config.undecided - eliminatee).size() > 0)){
                     pm::Map<Int, pm::Rational> coeffmap;
                     pm::Rational coeff(c[eliminatee.front()]/c[elim]);
                     for(const auto& e: eliminatee){
                        coeffmap[e] = c[e]/c[elim];
                        flip_gkz_vectors.row(e) -= coeffmap[e] * flip_gkz_vectors.row(elim);
                     }
                     Subsystem subconfig(config);
                     subconfig.active_rows -= elim;
                     subconfig.undecided -= eliminatee;
                     recursive_process(subconfig);
                     update_config(config, subconfig);
                     for(const auto& e: eliminatee){
                        flip_gkz_vectors.row(e) += coeffmap[e] * flip_gkz_vectors.row(elim);
                     }

                  }
               } else if(positives.size() == 0 || negatives.size() == 0){
                  pm::Set<Int> face = positives.size() == 0 ? negatives : positives;
                  face = config.active_rows - face;
                  if((config.undecided * face).size() > 0){
                     Subsystem subconfig(config);
                     subconfig.active_rows *= face;
                     subconfig.undecided *= face;
                     recursive_process(subconfig);
                     update_config(config, subconfig);
                  }
               }
            }

            i++;
         }
         // pm::cerr << "Matrix after presolve:" << std::endl;
         // pm::cerr << "Undecided: " << config.undecided << " Regular: " << config.rays << std::endl;
         // pm::cerr << pm::Matrix<pm::Rational>(flip_gkz_vectors.minor(config.active_rows, pm::All)) << std::endl;
         while(config.undecided.size() > 0){
            Int index = config.undecided.front();
            bool check = solve_via_LP(index, config);
            config.undecided -= index;
            if(check) config.rays += index;
            else {
               config.non_rays += index;
               config.active_rows -= index;
               break;
            }
         }
      }


      bool solve_via_LP(Int i, Subsystem& config){
         lp_necessary++;
         if(!config.undecided.contains(i)){
            throw std::runtime_error("Wrong input!");
         }
         pm::Set<Int> good(config.active_rows - i);
         pm::Matrix<pm::Rational> fgkzvs(flip_gkz_vectors.minor(config.active_rows - i, pm::All));
         fgkzvs = -flip_gkz_vectors.row(i) / fgkzvs;
         
         pm::Matrix<pm::Rational> lp_input(T(fgkzvs));
         Int nrows = lp_input.cols();
         const pm::Vector<pm::Rational> Obj(Scalar(0) | pm::ones_vector<pm::Rational>(nrows-1));
         const bool maximize = false;
         static const polymake::polytope::LP_Solver<pm::Rational>& solver = polymake::polytope::get_LP_solver<pm::Rational>();
         const auto S = solver.solve(pm::unit_matrix<pm::Rational>(nrows), lp_input, Obj, maximize);
         bool result;
         switch (S.status) {
            case polymake::polytope::LP_status::valid:
            case polymake::polytope::LP_status::unbounded:
               return false;
            default:
               return true;
         }
      }
      

      void init(){
         lp_necessary = 0;
         flip_gkz_vectors = pm::SparseMatrix<pm::Rational>(fimap.size(), source.no());
         Int i = 0;
         for(auto& [k,v] : fimap)
         {
            if(!v.isValid){
               irrelevant += i;
            } else {
               flip_gkz_vectors.row(i) = get_gkz_inner(k, v);
            }
            i++;
         }
      }


   public:
      HeuristicSolver(const TriangNode& s, const gkz_type& gkz_in, std::unordered_map<Flip, FlipInfoType>& fimap_in): 
         source(s),
         gkz(gkz_in),
         fimap(fimap_in){
            init();
      }
      
      const pm::Set<Int>& get_non_regular() const {
         return non_regular;
      }

      const pm::Set<Int>& get_regular() const {
         return regular;
      }

      const pm::Set<Int>& get_trivially_regular() const {
         return irrelevant;
      }
      
      void process() {
         Subsystem config;
         config.active_rows = pm::sequence(0, flip_gkz_vectors.rows()) - irrelevant;
         config.rays = regular;
         config.non_rays = non_regular;
         config.undecided = config.active_rows - regular - non_regular - no_lp_necessary;
         while(config.undecided.size() > 0){
            recursive_process(config);
         }
         regular = config.rays;
         non_regular = config.non_rays;
      }

      void mark_flip(Int i, bool isRegular){
         // undecided -= i;
         if(isRegular){
            regular += i;
         } else {
            non_regular += i;
            irrelevant += i;
            // prune();
         }
      }

      void do_not_consider(Int i){
         no_lp_necessary += i;
      }

      Int no_of_lp_calls() const {
         return lp_necessary;
      }
};


class RegularityCheckCache {
   private:
      mutable LRUCache<TriangNode, bool> cache;
      mutable LRUCache<std::pair<TriangNode, Flip>, bool> regularFlipCache;
      const PointConfiguration& pts;
      const Chirotope& chiro;
      pm::Matrix<Scalar> pmPts;
      const gkz_type& gkz;
      mutable Int lp_counter;


      void init(const std::string& lp_client);

   public: 
      RegularityCheckCache(Int size, const PointConfiguration& pts_in, const Chirotope& chiro_in, const std::string& lp_client, const gkz_type& gkz_in):
         pts(pts_in), chiro(chiro_in), cache(size), pmPts(pts.no(), pts.rank()), gkz(gkz_in),
         regularFlipCache(size)
         {
            lp_counter = 0;
            init(lp_client);
            // Convert topcom pts to polymake pts.
            pm::Rational x;
            for(Int i=0; i<pts.no(); i++){
               for(Int j=0; j<pts.rank(); j++){
                  x.copy_from(pts[i][j].get_mpq_t());
                  pmPts(i,j) = pm::convert_to<Scalar>(x);
               }
            }
         };

      bool isRegular(const TriangNode& t) const{
         bool result;
         if(cache.exist(t)){
            result = cache.get(t);
         } else {
            lp_counter++;
            result = isRegular_inner(t);
            // analyse_sec_cone(t);
            cache.put(t, result);
         }
         return result;
      }



      template<typename FlipInfoType, typename checkFine>
      void mark_regular_flips(const TriangNode& source, std::unordered_map<Flip, FlipInfoType>& all_flips) const {
         HeuristicSolver hs(source, gkz, all_flips);
         Int i = 0;
         for(auto& [k,v] : all_flips)
         {
            std::pair<TriangNode, Flip> key(source, k);
            if(regularFlipCache.exist(key)){
               hs.mark_flip(i, regularFlipCache.get(key));
            }
            if(checkFine::value){
               if(!v.isFine) hs.do_not_consider(i);
            }
            i++;
         }
         hs.process();
         lp_counter += hs.no_of_lp_calls();
         const pm::Set<Int>& regular(hs.get_regular()), non_regular(hs.get_non_regular());
         i = 0;
         for(auto& [k,v] : all_flips)
         {
            if(v.isValid){
               v.isValid = regular.contains(i);
               if(v.isValid){
                  store_flip(source, k, v, true);
               }
               if(non_regular.contains(i)){
                  store_flip(source, k, v, false);
               }
            }
            i++;
         }
      }

      void print_stat() const {
         std::cerr << "Solved " << lp_counter << " linear programs." << std::endl;
         std::cerr << "Stats of regular triangulations cache: ";
         cache.print_stat();
         std::cerr << "Stats of regular flips cache: ";
         regularFlipCache.print_stat();
      }

      void store_flip(const TriangNode& s, const Flip& flip, const FlipInfoSym& fi, bool result) const {
         std::pair<TriangNode, Flip> key(s, flip);
         Flip reverseFlip(flip);
         reverseFlip.swap();
         const Symmetry& g(fi.optimizer);
         std::pair<TriangNode, Flip> reverseKey(fi.target, g(reverseFlip));
         regularFlipCache.put(key, result);
         regularFlipCache.put(reverseKey, result);
      }
      
      void store_flip(const TriangNode& s, const Flip& flip, const FlipInfo& fi, bool result) const {
         std::pair<TriangNode, Flip> key(s, flip);
         Flip reverseFlip(flip);
         reverseFlip.swap();
         std::pair<TriangNode, Flip> reverseKey(fi.target, reverseFlip);
         regularFlipCache.put(key, result);
         regularFlipCache.put(reverseKey, result);
      }
     

      void analyse_sec_cone(const TriangNode& tn) const {
         TriangFlips TF(chiro, tn, SymmetryGroup(chiro.no()), false);
         std::cerr << "Testing:" << std::endl;
         std::cerr << tn << std::endl;
         // std::cerr << TF << std::endl;
         const MarkedFlips& mf(TF.flips());
         Int nflips = 0;
         for(MarkedFlips::const_iterator flip = mf.begin(); flip!=mf.end(); flip++) nflips++;
         pm::Matrix<pm::Rational> flipineq(nflips, pts.no());
         std::cerr << mf.size() << " " << pts.no() << std::endl;
         Int i = 0;
         for(MarkedFlips::const_iterator flip = mf.begin(); flip!=mf.end(); flip++){
            const auto FR = flip->key();
            Flip F(tn, FR);
            // std::cerr << gkz.gkz_as_string(F) << std::endl;
            flipineq.row(i) = gkz.evaluate(F);
            i++;
            // TriangNode Ft(0, tn, F);
            // std::cerr << Ft << std::endl;
         }
         std::cerr << "Total flips: " << i << std::endl;
         // pm::cout << flipineq << pm::endl;
         static const pm::perl::OptionSet noopts;
         pm::Array<pm::Set<Int>> pmMC = triangulation2polymake(tn);
         std::pair<const pm::Matrix<Scalar>, const pm::Matrix<Scalar>> SC = polymake::polytope::secondary_cone_ineq(pmPts, pmMC, noopts);
         polymake::BigObject polymake_cone("polytope::Cone<Rational>");
         polymake_cone.take("INEQUALITIES") << SC.first;
         polymake_cone.take("EQUATIONS") << SC.second;
         polymake::BigObject our_cone("polytope::Cone<Rational>");
         our_cone.take("INEQUALITIES") << flipineq;

         Int polymake_dim = polymake_cone.give("CONE_DIM");
         Int our_dim = our_cone.give("CONE_DIM");
         std::cerr << "Dim check: " << (polymake_dim == our_dim) << std::endl;
         

         std::cerr << std::endl;
      }

      std::string get_secondary_cone_rays(const TriangNode& t) const {
         // Convert triangulation to polymake datatype
         pm::Array<pm::Set<Int>> pmMC = triangulation2polymake(t);
         // Make secondary cone ineqs
         static const pm::perl::OptionSet pos("lift_face_to_zero", 0);
         std::pair<const pm::Matrix<Scalar>, const pm::Matrix<Scalar>> SC = polymake::polytope::secondary_cone_ineq(pmPts, pmMC, pos);
         polymake::BigObject cone("polytope::Cone<Rational>");
         cone.take("INEQUALITIES") << SC.first;
         cone.take("EQUATIONS") << SC.second;
         pm::Matrix<pm::Rational> rays = cone.give("RAYS");
         std::string result = "Secondary cone rays: [";
         Int nrows = rays.rows();
         Int ncols = rays.cols();
         Int i = 0;
         Int j;
         for(const auto& row : rows(rays)){
            result += "[";
            j = 0;
            for(const auto& entry : row){
               result += scalar2string(entry);
               j++;
               if(j < ncols) result += ",";
            }
            result += "]";
            i++;
            if(i < nrows) result += ",";
         }
         return result;
      }

   
      bool subdivision_is_regular(const pm::Array<pm::Set<Int>>& pmMC) const {
         // Make secondary cone ineqs
         static const pm::perl::OptionSet pos("lift_face_to_zero", 0);
         std::pair<const pm::Matrix<Scalar>, const pm::Matrix<Scalar>> SC = polymake::polytope::secondary_cone_ineq(pmPts, pmMC, pos);
         bool result = is_regular_lp(SC);
         return result;
      }

   private:
      pm::Array<pm::Set<Int>> triangulation2polymake(const TriangNode& t) const {
         pm::Array<pm::Set<Int>> pmMC(t.card());
         Int i = 0;
         for(auto s : t) {
            pmMC[i] = pm::convert_to<pm::Set<Int>>(s);
            i++;
         }
         return pmMC;
      }

      bool isRegular_inner(const TriangNode& t) const{

         // Convert triangulation to polymake datatype
         pm::Array<pm::Set<Int>> pmMC = triangulation2polymake(t);
         // Make secondary cone ineqs
         static const pm::perl::OptionSet pos("lift_face_to_zero", 0);
         std::pair<const pm::Matrix<Scalar>, const pm::Matrix<Scalar>> SC = polymake::polytope::secondary_cone_ineq(pmPts, pmMC, pos);
         bool result = is_regular_lp(SC);
         return subdivision_is_regular(pmMC);
      }

      bool is_regular_lp(std::pair<const pm::Matrix<Scalar>, const pm::Matrix<Scalar>> SC) const {
         const Int n_vertices=pts.no();
         pm::Matrix<pm::Rational> H(SC.first);
         pm::Matrix<pm::Rational> E(SC.second);
         const pm::Vector<pm::Rational> Obj(Scalar(0) | pm::ones_vector<Scalar>(n_vertices));
         bool maximize = false;
         E = pm::zero_matrix<pm::Rational>(E.rows(),1) | E;
         H = ((pm::Rational(-1))*(pm::ones_matrix<pm::Rational>(H.rows(),1))) | H;

         static const polymake::polytope::LP_Solver<pm::Rational>& solver = polymake::polytope::get_LP_solver<pm::Rational>();
         const auto S = solver.solve(H, E, Obj, maximize);
         switch (S.status) {
            case polymake::polytope::LP_status::valid:
            case polymake::polytope::LP_status::unbounded:
               return true;
            default:
               return false;
         }
         return false;
      }

};



